#!/bin/sh
# the next line restarts using tclsh \
exec tclsh "$0" ${1+"$@"}

################################################################################
#                                                                              #
#  Copyright 2014                                                              #
#                                                                              #
# This software is copyrighted by Gerald W. Lester. The following terms        #
# apply to all files associated with the software unless explicitly disclaimed #
# in individual files.                                                         #
#                                                                              #
# The authors hereby grant permission to use, copy, modify, distribute,        #
# and license this software and its documentation for any purpose, provided    #
# that existing copyright notices are retained in all copies and that this     #
# notice is included verbatim in any distributions. No written agreement,      #
# license, or royalty fee is required for any of the authorized uses.          #
# Modifications to this software may be copyrighted by their authors           #
# and need not follow the licensing terms described here, provided that        #
# the new terms are clearly indicated on the first page of each file where     #
# they apply.                                                                  #
#                                                                              #
# IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY         #
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES          #
# ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY           #
# DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE            #
# POSSIBILITY OF SUCH DAMAGE.                                                  #
#                                                                              #
# THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,           #
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,    #
# FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE       #
# IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE       #
# NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR     #
# MODIFICATIONS.                                                               #
#                                                                              #
# GOVERNMENT USE: If you are acquiring this software on behalf of the          #
# U.S. government, the Government shall have only "Restricted Rights"          #
# in the software and related documentation as defined in the Federal          #
# Acquisition Regulations (FARs) in Clause 52.227.19 (c) (2).  If you          #
# are acquiring the software on behalf of the Department of Defense, the       #
# software shall be classified as "Commercial Computer Software" and the       #
# Government shall have only "Restricted Rights" as defined in Clause          #
# 252.227-7014 (b) (3) of DFARs.  Notwithstanding the foregoing, the           #
# authors grant the U.S. Government and others acting in its behalf            #
# permission to use and distribute the software in accordance with the         #
# terms specified in this license.                                             #
#                                                                              #
################################################################################

package require Tcl 8.6
##
## Please see the Protocol.rtf
##

##
## Set the following to enable debug logging.  Most log entries, except for
## data received, begin with a "#" -- thus the log file can be feed back
## into the bot to "replay" the game for debugging (assuming the bot makes
## deterministic moves).
##      io      -- log what is recevied from and sent to the game engine.
##      trace   -- log entry and exits from the procedures.
##      debug   -- Reserved for your debugging information.  You can also use
##                  any other "level" words you want.
##
set Settings(DebugEnabled) {}
set Settings(DebugEnabled) {io debug trace}

##
## Replace the following procedures in the ::bot:: namespace with the actual
## implementation for the bot.
##
namespace eval ::bot:: {
    array set strategy {}
    set exteriorRegionList {}
    set interiorRegionList {}
    set weightedMoveList {}

    ##
    ## This procedure is used to initialize your custom code. It is run after
    ## all of the global variables have been initialized but prior to starting
    ## to get commands from the game engine. This is where you should put any
    ## expensive intiialziation code (like spawning threads).
    ##
    proc DoInitialization {} {
        Log trace "Entering [info level 0]"
        variable strategy

        array set strategy {
            expansionWeight 10
            offenseWeight   10
            defenseWeight   500
            minAttackRatio  3.0
            defenseRatio    1.25
        }

        return;
    }

    proc GetDistanceToExterior {fromRegion path} {
        Log trace "Entering [info level 0]"

        variable distanceArray
        variable exteriorRegionList
        global Map

        if {[info exists distanceArray($fromRegion,distance)]} {
            set distance $distanceArray($fromRegion,distance)
        } elseif {$fromRegion in $path} {
            return 999999999
        } else {
            set distance 999999999
            foreach neighbor $Map($fromRegion,neighbors) {
                if {$neighbor in $exteriorRegionList} {
                    set distance 0
                    set toRegion $neighbor
                    break
                } else {
                    set temp [GetDistanceToExterior $neighbor [concat $path $fromRegion]]
                    if {$temp <  $distance} {
                        set toRegion $neighbor
                        set distance $temp
                    }
                }
            }
            if {$distance != 999999999} {
                incr distance
                set distanceArray($fromRegion,toPath) $toRegion
                set distanceArray($fromRegion,distance) $distance
            }
        }
        Log trace "[lindex [info level 0] 0] -> {$distance}"
        return $distance
    }

    ##
    ## Get and weight list of possible moves
    ##
    proc GetWeightedMoveList {} {
        Log trace "Entering [info level 0]"

        variable strategy
        variable exteriorRegionList
        variable interiorRegionList
        variable weightedMoveList
        variable distanceArray
        global Map
        global Bonuses
        global Settings

        set weightedMoveList [list]
        GetRegionTypes

        ##
        ## First off process the interior regions
        ##
        foreach region $interiorRegionList {
            if {$Map($region,armyCount) > 1} {
                GetDistanceToExterior $region [list]
                ##
                ## Attempt to move the nearest external region, and make that the highest priority
                ##
                set tmp [expr {$Map($region,armyCount) - 1}]
                lappend weightedMoveList [list 999999999 M $region $distanceArray($region,toPath) $tmp]
            }
        }

        ##
        ## Compute our bonus weights
        ##
        array set bonusWeights {}
        foreach key [array names Bonuses *,value] {
            set bonus [lindex [split $key {,}] 0]
            set bonusWeights($bonus) 1
        }
        foreach region [concat $exteriorRegionList $interiorRegionList] {
            incr bonusWeights($Map($region,bonus))
        }
        foreach {bonus count} [array get bonusWeights] {
            set bonusWeights($bonus) [expr {$Bonuses($bonus,value) - ([llength $Bonuses($bonus,contains)] - (1.0 * $count) )} ]
        }

        ##
        ## Ok, now lets process the exterior regions
        ##
        array set weightArray {}
        array set defenseArray {}
        array set defenseCountArray {}
        foreach region $exteriorRegionList {
            set defenseArray($region) 0.0
            set defenseCountArray($region) 0
            foreach neighbor $Map($region,neighbors) {
                set weightArray($neighbor) 0.0
                set regionArray($neighbor) [list]
            }
        }

        foreach region $exteriorRegionList {
            foreach neighbor $Map($region,neighbors) {
                ##
                ## Compute the expansion
                ##
                set weight [expr {$strategy(expansionWeight) * $bonusWeights($Map($neighbor,bonus))}]
                if {$Map($neighbor,ownedBy) eq $Settings(MyName)} {
                    continue
                } else {
                    set weightArray($neighbor) [expr {$weightArray($neighbor) + $weight}]
                }
                lappend regionArray($neighbor) $region

                if {$Map($neighbor,ownedBy) eq "Neutral"} {
                    ##
                    ## Done if this is a neutral
                    ##
                    continue
                }

                ##
                ## Add in the offense weight if this is a player
                ##
                set weight $strategy(offenseWeight)
                set weightArray($neighbor) [expr {$weightArray($neighbor) + $weight}]

                ##
                ## Compute the defense
                ##
                set weight $strategy(defenseWeight)
                set count [expr {int(max(1,round($Map($neighbor,armyCount) * $strategy(defenseRatio))))}]
                set defenseArray($region) [expr {$defenseArray($region) + $weight}]
                set defenseCountArray($region) [expr {$defenseCountArray($region) + $count}]
            }
        }

        ##
        ## Add the attacks in
        ##
        foreach {neighbor weight} [array get weightArray] {
            if  {$Map($neighbor,ownedBy) eq "Neutral"} {
                set count [expr {$Map($neighbor,armyCount) * 2}]
            } else {
                set count [expr {int(max(1,round($Map($neighbor,armyCount) * $strategy(minAttackRatio))))}]
            }
            foreach region $regionArray($neighbor) {
                lappend weightedMoveList [list $weight a $region $neighbor $count]
            }
        }

        ##
        ## Add the defense in
        ##
        foreach {region weight} [array get defenseArray] {
            if {$weight > 0.01} {
                lappend weightedMoveList [list $weight d $region $neighbor $defenseCountArray($region)]
            }
        }

        if {$strategy(expansionWeight) > 10} {
            incr strategy(expansionWeight) -2
            incr strategy(offenseWeight)
            incr strategy(defenseWeight)
        } elseif {$strategy(expansionWeight)} {
            incr strategy(expansionWeight) -1
            incr strategy(defenseWeight) 1
        } else {
            incr strategy(offenseWeight)
            incr strategy(defenseWeight) -1
        }

        set weightedMoveList [lsort -decreasing -real -index 0 $weightedMoveList]
        Log debug "Weighted Move List is {$weightedMoveList}"
        Log trace "[lindex [info level 0] 0] -> {}"
        return
    }

    ##
    ## Get list of exterior and interior regions
    ##
    proc GetRegionTypes {} {
        Log trace "Entering [info level 0]"

        variable exteriorRegionList
        variable interiorRegionList

        ##
        ## Build the list of interior and exterior regions
        ##
        set exteriorRegionList [list]
        set interiorRegionList [list]
        foreach region $::OwnedRegionsList {
            set exterior no
            foreach neighbor $::Map($region,neighbors) {
                if {$::Map($neighbor,ownedBy) ne $::Settings(MyName)} {
                    set exterior yes
                }
            }
            if {$exterior} {
                lappend exteriorRegionList $region
            } elseif {$::Map($region,armyCount)  > 1} {
                lappend interiorRegionList $region
            }
        }

        Log trace "[lindex [info level 0] 0] -> {}"
        return
    }

    ##
    ## This procedure should return a list of PickCount regions that this
    ## bot would like
    ##
    proc DoPickStartingPositions {PickCount AvailableRegions} {
        Log trace "Entering [info level 0]"

        global Map
        global Bonuses

        set outList [list]

        ##
        ## We want to pick regions such that we maximize the
        ## owned/regions-in-bonuse ratio
        ##
        array set bonusCnt {}
        array set inBonus {}
        foreach region $AvailableRegions {
            set bonus $Map($region,bonus)
            incr bonusCnt($bonus)
            lappend inBonus($bonus) $region
        }
        set bonusList [list]
        foreach {bonus count} [array get bonusCnt] {
            lappend bonusList [list $bonus [expr {(1.0 * $count)/[llength $Bonuses($bonus,contains)]}]]
        }
        set bonusList [lsort -index 1 -decreasing -real $bonusList]
        foreach {bonusInfo} $bonusList {
            lassign $bonusInfo bonus rating
            lappend outList {*}$inBonus($bonus)
        }

        ##
        ## And return the correct number of picks
        ##
        set result [lrange $outList 0 [expr {$PickCount - 1}]]
        Log trace "[lindex [info level 0] 0] -> {$result}"
        return $result
    }

    ##
    ## This returns a list of duples consiting of region and number of armies
    ##
    proc DoPlaceArmies {} {
        Log trace "Entering [info level 0]"

        variable strategy
        variable exteriorRegionList
        variable interiorRegionList
        variable weightedMoveList

        global Map
        array set tmpMap [array get Map]

        GetWeightedMoveList

        set placementList [list]
        array set allocationArray {}
        set avaliableArmies $::Settings(AvailableArmies)

        ##
        ## Walk the Weighted movement list until we are out of armies
        ##
        foreach moveItem $weightedMoveList {
            lassign $moveItem weight type fromRegion toRegion count
            switch -exact $type {
                m {
                    ##
                    ## Movement, do nothing
                    ##
                }
                d {
                    ##
                    ## Defense, add arimes
                    ##
                    set tmp [expr {min($avaliableArmies,$count)}]
                    if {$tmp > 0} {
                        incr avaliableArmies -$tmp
                        incr allocationArray($fromRegion) $tmp
                    }
                }
                default {
                    ##
                    ## Attack
                    ##
                    if {$tmpMap($fromRegion,armyCount) > 1} {
                        set hasNow $tmpMap($fromRegion,armyCount)
                        incr hasNow -1
                        set tmpMap($fromRegion,armyCount) [expr {max(0,$hasNow - $count)}]
                        set count [expr {max(0,$count - $hasNow)}]
                    }
                    set tmp [expr {min($avaliableArmies,$count)}]
                    if {$tmp > 0} {
                        incr avaliableArmies -$tmp
                        incr allocationArray($fromRegion) $tmp
                    }
                }
            }
            if {$avaliableArmies < 1} {
                break
            }
        }

        ##
        ## Distribute the remainders
        ##
        while {$avaliableArmies > 0} {
            foreach region $exteriorRegionList {
                incr allocationArray($region)
                incr avaliableArmies -1
                if {$avaliableArmies < 1} {
                    break
                }
            }
        }

        ##
        ## Build the placement list
        ##
        foreach {region value}  [array get allocationArray] {
            if {$value > 0} {
                lappend placementList [list $region $value]
                incr ::Map($region,armyCount) $value
            }
        }

        ##
        ## And return the placements
        ##
        Log trace "[lindex [info level 0] 0] -> {$placementList}"
        return $placementList
    }

    ##
    ## This returns a list of attack/transfer moves in the form of triplets
    ## consisting of: source region, target region and number of armies.
    ##
    proc DoMovements {} {
        Log trace "Entering [info level 0]"

        variable weightedMoveList
        global Map

        set movementList [list]

        ##
        ## Do the movements
        ##
        foreach moveItem $weightedMoveList {
            lassign $moveItem weight type fromRegion toRegion attackCount
            if {($type eq "d")} {
                continue
            }
            incr Map($fromRegion,armyCount) -$attackCount
            lappend movementList [list $fromRegion $toRegion $attackCount]
        }

        ##
        ## And return the movement list
        ##
        Log trace "[lindex [info level 0] 0] -> {$movementList}"
        return $movementList
    }

    namespace export Do*
}


###############################################################################
##                                                                           ##
##                      DO NOT MODIFY BELOW THIS LINE                        ##
##                                                                           ##
###############################################################################

##
## The history is an array indexed by turns that contains a list of opponent
## movements.
##
array set History {}

##
## List of the regions owned by this bot
##
set OwnedRegionsList [list]

##
## The map will consist of two arrays, Map and Bonuses.
##
## The indexes for Bonuses are:
##      <bonus>,value       - contains the value of the bonus
##      <bonus>,contains    - contains a list of regions in the bonus
##
## The indexes for Map are:
##      <region>,bonus      - which bonus, if any this region is part of
##      <region>,neighbors  - the regions this bonus connects to
##      <region>,lastViewed - last turn this was viewed
##      <region>,ownedBy    - player owning this last time it was viewed
##      <region>,armyCount  - how many armies were in it when lasted viewed
##
array set Bonuses {}
array set Map {}

##
## The indexes for Settings are:
##  Turn                - the current turn
##  AvailableArmies     - how many armies are available to be allocated
##  MyName              - this bot's name
##  Opponents           - a list of opponents
##  DropDeadTimestamp   - the time, in milliseconds, when this command will time
##                        out
##  ThreadsAvailable    - a flag is available if the thread package is available
##  MaxThreadCount      - Maximun number of threads that should be spawned
##  DebugEnabled        - Boolean indicating if debug information on what is
##                        sent and received is logged to stderr
##
array set Settings {
    Turn 0
    AvailableArmies 0
    MyName {}
    Opponents {}
    DropDeadTimestamp {}
}

##
## Figure out if the bot can be multi threaded
##
if {$tcl_platform(threaded) && [catch {package require Thread}]} {
    set Settings(ThreadsAvailable) no
    set Settings(MaxThreadCount) 0
} else {
    set Settings(ThreadsAvailable) yes
    switch -exact -- $tcl_platform(os) {
        Windows {
            set Settings(MaxThreadCount) $env(NUMBER_OF_PROCESSORS)
        }
        Linux {
            set Settings(MaxThreadCount) 1
            if {[catch {
                set fh [open "/proc/cpuinfo" r]
                while {[gets $fh line] >= 0} {
                    if {[string match processor* $line]} {
                        incr Settings(MaxThreadCount)
                    }
                }
                close $fh
            }]} {
                catch {
                    set rawCount [exec sysctl hw.ncpu]
                    set Settings(MaxThreadCount) [string trim [lindex [split $rawCount {:}] end]]
                }
            }
        }
        Darwin {
            set Settings(MaxThreadCount) 1
            catch {
                set rawCount [exec sysctl hw.ncpu]
                set Settings(MaxThreadCount) [string trim [lindex [split $rawCount {:}] end]]
            }
        }
        default {
            set Settings(MaxThreadCount) 1
        }
    }
    incr Settings(MaxThreadCount) -1
    if {$Settings(MaxThreadCount) < 1} {
        set Settings(MaxThreadCount) 0
    }
}

##
## Below this line is the code that implements the basic bot
##
namespace eval ::setup_map:: {

    ##
    ##  The superregions are given, with their bonus armies reward, all
    ##  separated by spaces. Odd numbers are superregion ids, even numbers are
    ##  rewards.
    ##
    proc super_regions {args} {
        Log trace "Entering [info level 0]"

        foreach {region value} $args {
            set ::Bonuses($region,value) $value
        }
        Log trace "[lindex [info level 0] 0] -> {}"
        return;
    }

    ##
    ##  The regions are given, with their parent superregion, all separated by
    ##  spaces. Odd numbers are the region ids, even numbers are the superregion
    ##  ids.
    ##
    proc regions {args} {
        Log trace "Entering [info level 0]"

        foreach {region bonus} $args {
            set ::Map($region,bonus) $bonus
            lappend ::Bonuses($region,contains) $region
        }
        Log trace "[lindex [info level 0] 0] -> {}"
        return;
    }

    ##
    ##  The connectivity of the regions are given, first is the region id. Then
    ##  the neighbouring regions' ids, separated by commas. Connectivity is only
    ##  given in one way: 'region id' < 'neighbour id'.
    ##
    proc neighbors {args} {
        Log trace "Entering [info level 0]"

        set changedList [list]
        foreach  {region neighbors} $args {
            lappend changedList $region
            foreach neighbor [split $neighbors {,}] {
                lappend ::Map($region,neighbors) $neighbor
                lappend ::Map($neighbor,neighbors) $region
                lappend changedList $neighbor
            }
        }
        foreach region [lsort -integer -unique $changedList] {
            set ::Map($region,neighbors) [lsort -integer -unique $::Map($region,neighbors) ]
        }
        Log trace "[lindex [info level 0] 0] -> {}"
        return;
    }

    namespace export super_regions regions neighbors
    namespace ensemble create
}

namespace eval ::settings:: {

    ##
    ##  The name of your bot is given.
    ##
    proc your_bot {name} {
        Log trace "Entering [info level 0]"

        set ::Settings(MyName) $name
        Log trace "[lindex [info level 0] 0] -> {}"
        return;
    }

    ##
    ##  The name of your opponent bot is given.
    ##
    proc opponent_bot {name} {
        Log trace "Entering [info level 0]"

        lappend ::Settings(Opponents) $name
        Log trace "[lindex [info level 0] 0] -> {}"
        return;
    }

    ##
    ##  The amount of armies your bot can place on the map at the start of this
    ##  round.
    ##
    proc starting_armies {armyCount} {
        Log trace "Entering [info level 0]"

        set ::Settings(AvailableArmies) $armyCount
        incr ::Settings(Turn)
        Log trace "[lindex [info level 0] 0] -> {}"
        return;
    }

    namespace export your_bot opponent_bot starting_armies
    namespace ensemble create
}

namespace eval ::go:: {

    ##
    ##  Request for the bot to return his place armies moves.
    ##
    proc place_armies {timeLimit} {
        Log trace "Entering [info level 0]"

        set ::Settings(DropDeadTimestamp) [expr {[clock milliseconds] + $timeLimit}]
        set placementList [::bot::DoPlaceArmies]
        set resultList [list]
        foreach placement $placementList {
            lappend resultList "$::Settings(MyName) place_armies $placement"
        }
        set results [join $resultList {, }]
        Log trace "[lindex [info level 0] 0] -> {$results}"
        return $results
    }

    ##
    ##  Request for the bot to return his attack and/or transfer moves
    ##  engine-output.
    ##
    proc attack/transfer {timeLimit} {
        Log trace "Entering [info level 0]"

        set ::Settings(DropDeadTimestamp) [expr {[clock milliseconds] + $timeLimit}]
        set movementList [::bot::DoMovements]
        set results {}
        if {[llength $movementList]} {
            set resultList {}
            foreach movment $movementList {
                lappend resultList "$::Settings(MyName) attack/transfer $movment"
            }
            set results [join $resultList {, }]
        } else {
            set results {No moves}
        }
        Log trace "[lindex [info level 0] 0] -> {$results}"
        return $results
    }

    namespace export place_armies attack/transfer
    namespace ensemble create
}

##
##  Starting regions to be chosen from are given, request to return 6 region
##  ids.
##
proc pick_starting_regions {timeLimit args} {
    Log trace "Entering [info level 0]"

    set ::Settings(DropDeadTimestamp) [expr {[clock milliseconds] + $timeLimit}]
    set pickList [::bot::DoPickStartingPositions 6 $args]
    Log trace "[lindex [info level 0] 0] -> {$pickList}"
    return $pickList
}

##
##  Visible map for the bot is given like this: region id; player owning region; number of armies.
##
proc update_map {args} {
    Log trace "Entering [info level 0]"

    set ::OwnedRegionsList [list]
    foreach {region owner armies} $args {
        set ::Map($region,lastViewed) $::Settings(Turn)
        set ::Map($region,ownedBy) $owner
        set ::Map($region,armyCount) $armies
        if {$owner eq $::Settings(MyName)} {
            lappend ::OwnedRegionsList $region
        }
    }
    Log trace "[lindex [info level 0] 0] -> {}"
    return;
}

##
##  All the visible moves the opponent has done are given in consecutive order.
##  Can be any move and has the same format as in the table below.
##
proc opponent_moves {args} {
    Log trace "Entering [info level 0]"

    foreach {opponent source target armies} $args {
        lappend ::History($::Settings(Turn)) [list opponent $source $target $armies]
    }

    Log trace "[lindex [info level 0] 0] -> {}"
    return;
}

##
## debug logging procedure
##
proc Log {level txt} {
    if {$level in $::Settings(DebugEnabled)} {
        chan puts stderr "# [clock format [clock seconds]] $txt"
        chan flush stderr
    }
}

##
## Initialize the custom part of the bot
##
::bot::DoInitialization

##
## See if we are to put out prompts, and if so put out the initial one.
##
if {$argc == 1} {
    set prompt [lindex $argv 0]
    set usePrompt yes
    chan puts -nonewline stdout "Turn $::Settings(Turn) $prompt"
    chan flush stdout
} else {
    set usePrompt no
}

##
## Evaluate commands from the game engine
##
while {[chan gets stdin cmd] != -1} {
    Log debug "Start of Turn $::Settings(Turn)"
    Log io "received: \n$cmd"
    if {[catch {eval $cmd} msg]} {
        chan puts stderr $::errorInfo
        chan flush stderr
    } elseif {[string length $msg]} {
        chan puts stdout $msg
        flush stdout
        Log io " replied: $msg"
    } else {
        Log io "\t no reply needed"
    }
    Log debug "End of Turn $::Settings(Turn)"
    if {$usePrompt} {
        chan puts -nonewline stdout "Turn $::Settings(Turn) $prompt"
        chan flush stdout
    }
}

##
## All done
##
exit 0
