#!/bin/sh
# the next line restarts using tclsh \
exec tclsh "$0" ${1+"$@"}

##
## Usage:
##      GameName                - name of game
##      GameBaseDirectory       - path (absolute or relative to current directory)
##      MapSpecificationFile    - Specification file (absolute or relative to
##                                GameBaseDirectory) for map
##      <playerInfo> -- consists of pairs of argument
##          PlayerName      - name of player
##          PlayerBotFile   - path (absolute or relative to GameBaseDirectory)
##                            for bot code

################################################################################
#                                                                              #
#  Copyright 2014                                                              #
#                                                                              #
# This software is copyrighted by the Gerald W. Lester. The following terms    #
# apply to all files associated with the software unless explicitly disclaimed #
# in individual files.                                                         #
#                                                                              #
# The authors hereby grant permission to use, copy, modify, distribute,        #
# and license this software and its documentation for any purpose, provided    #
# that existing copyright notices are retained in all copies and that this     #
# notice is included verbatim in any distributions. No written agreement,      #
# license, or royalty fee is required for any of the authorized uses.          #
# Modifications to this software may be copyrighted by their authors           #
# and need not follow the licensing terms described here, provided that        #
# the new terms are clearly indicated on the first page of each file where     #
# they apply.                                                                  #
#                                                                              #
# IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY         #
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES          #
# ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY           #
# DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE            #
# POSSIBILITY OF SUCH DAMAGE.                                                  #
#                                                                              #
# THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,           #
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,    #
# FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE       #
# IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE       #
# NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR     #
# MODIFICATIONS.                                                               #
#                                                                              #
# GOVERNMENT USE: If you are acquiring this software on behalf of the          #
# U.S. government, the Government shall have only "Restricted Rights"          #
# in the software and related documentation as defined in the Federal          #
# Acquisition Regulations (FARs) in Clause 52.227.19 (c) (2).  If you          #
# are acquiring the software on behalf of the Department of Defense, the       #
# software shall be classified as "Commercial Computer Software" and the       #
# Government shall have only "Restricted Rights" as defined in Clause          #
# 252.227-7014 (b) (3) of DFARs.  Notwithstanding the foregoing, the           #
# authors grant the U.S. Government and others acting in its behalf            #
# permission to use and distribute the software in accordance with the         #
# terms specified in this license.                                             #
#                                                                              #
################################################################################

package require Tcl 8.6
##
## Please see the Protocol.rtf
##

##
## The map will consist of two arrays, Map and Bonuses.
##
## The indexes for Bonuses are:
##      <bonus>,name        - name of the bonus
##      <bonus>,value       - contains the value of the bonus
##      <bonus>,contains    - contains a list of regions in the bonus
##
## The indexes for Map are:
##      <region>,name           - name of the region
##      <region>,bonus          - which bonus, if any this region is part of
##      <region>,neighbors      - the regions this region connects to
##      <region>,botNeighbors   - the regions this region connects to
#
#		Difference neighbours vs botNeighbours ?
#
##      <region>,preAttackCount - armies in region after army placement
##      <region>,ownedBy        - player owning this region
##      <region>,armyCount      - how many armies are currently in it
##
## The indexes for History are:
##      <turn>,<player>         - list of moves observed by the player
##
array set Bonuses {}
array set Map {}
array set History {}
set MapFile {}
set GameName {}
set GameBaseDirectory [pwd]

##
## The indexes for Settings are:
##  Turn            - the current turn
##  MaxTurn         - the maximum number of turns
##  TurnMode        - how to process turns, must be either Serial or Parrallel (Serial)
##  BaseArmies      - base army allocation
##  Timeout         - maxium time a bot can take to respond
##  BonusList       - a list of regions
##  RegionList      - a list of regions
##  PlayerList      - a list of players
##  <bot>,channel   - communication channel to the bot
##
array set Settings {
    Turn 0
    MaxTurn 100
    TurnMode Serial
    BaseArmies 5
    Timeout 5000
    BonusList {}
    RegionList {}
    PlayerList {}
}

##
## Namespace that implements the map -- this is a superset of what gets sent to
## the players
##
namespace eval ::setup_map:: {

    ##
    ##  The superregions are given, with their bonus armies reward, all
    ##  separated by spaces. Odd numbers are superregion ids, even numbers are
    ##  rewards.
    ##
    proc super_regions {args} {
        foreach {region value} $args {
            set ::Bonuses($region,value) $value
            lappend ::Settings(BonusList) $region
        }
        return;
    }

    ##
    ##  The regions are given, with their parent superregion, all separated by
    ##  spaces. Odd numbers are the region ids, even numbers are the superregion
    ##  ids.
    ##
    proc regions {args} {
        foreach {region bonus} $args {
            set ::Map($region,bonus) $bonus
            lappend ::Bonuses($bonus,contains) $region
            lappend ::Settings(RegionList) $region
        }
        return;
    }

    ##
    ##  The connectivity of the regions are given, first is the region id. Then
    ##  the neighbouring regions' ids, separated by commas. Connectivity is only
    ##  given in one way: 'region id' < 'neighbour id'.
    ##
    proc neighbors {args} {
        set changedList [list]
        foreach  {region neighbors} $args {
            foreach neighbor [split $neighbors {,}] {
                if {![info exists ::Map($region,neighbors)] ||
		    ($neighbor ni $::Map($region,neighbors))} {
                    lappend ::Map($region,neighbors) $neighbor
                    lappend changedList $region
                }
                if {![info exists ::Map($neighbor,neighbors)] ||
		    ($region ni $::Map($neighbor,neighbors))} {
                    lappend ::Map($neighbor,neighbors) $region
                    lappend changedList $neighbor
                }
            }
        }
        foreach region [lsort -integer -unique $changedList] {
            set ::Map($region,neighbors) [lsort -integer -unique $::Map($region,neighbors) ]
        }
        return;
    }

    ##
    ## Graphic file for this map
    ##
    proc graphic {fileName} {
        set ::MapFile $fileName
    }

    ##
    ## Define a bonus (aka super region)
    ##
    proc define_super_region {bonus name} {
        set ::Bonuses($bonus,name) $name
    }

    ##
    ## Define a region
    ##
    proc define_region {region name} {
        set ::Map($region,name) $name
    }

    namespace export super_regions regions neighbors graphic define_super_region define_region
    namespace ensemble create
}

##
## This procedure will randomize a list
##
proc RandomizeList {OrderedList} {
    set outList [list]

    ##
    ## for the sample we will randomly pick
    ##
    set inList [list]
    foreach region $OrderedList {
        lappend inList [list [expr {rand()}] $region]
    }
    set inList [lsort -real -index 0 $inList]
    foreach pair $inList {
        lappend outList [lindex $pair 1]
    }

    return $outList
}

##
## Launch a player bot
##
proc LaunchBot {PlayerName PlayerBotFile} {
    set fileName $::GameName-$PlayerName.log
    set fn "|$PlayerBotFile 2> $fileName"
    set ::Settings($PlayerName,channel) [open $fn RDWR]
    set ::Settings($PlayerName,status) Opened
    chan configure $::Settings($PlayerName,channel) -buffering line
    chan event $::Settings($PlayerName,channel) readable [list PlayerReadable $PlayerName]
    lappend ::Settings(PlayerList) $PlayerName
    return
}

##
## Handle player data present
##
proc PlayerReadable {PlayerName} {
    set chan $::Settings($PlayerName,channel)
    if {[chan eof $chan]} {
        set ::Settings($PlayerName,status) EOF
    } elseif {[gets $chan ::Settings($PlayerName,data)] == -1} {
        set ::Settings($PlayerName,status) EOF
    } else {
        set ::Settings($PlayerName,status) Ready
        Log "Reply from $PlayerName: {$::Settings($PlayerName,data)}"
    }
    if {$::Settings($PlayerName,status) eq "EOF"} {
        Log "EOF on $PlayerName"
        catch {chan event $::Settings($PlayerName,channel) readable {}}
        catch {chan close $::Settings($PlayerName,channel)}
        set index [lsearch -exact $::Settings(PlayerList) $PlayerName]
        if {$index != -1} {
            set ::Settings(PlayerList) [lreplace $::Settings(PlayerList) $index $index]
        }
    }
    return
}

##
## Write data to player
##
proc SendToPlayer {PlayerName data {expectAnswer no} {timeoutFactor 1}} {
    Log "Sending to $PlayerName: {$data} -- waiting $expectAnswer"
    if {$expectAnswer} {
        set status [catch {
        set id [after [expr {$timeoutFactor * $::Settings(Timeout)}] [list set ::Settings($PlayerName,status) Timeout]]
        set ::Settings($PlayerName,status) Waiting
        catch {chan puts  $::Settings($PlayerName,channel) $data}
        vwait ::Settings($PlayerName,status)
        after cancel $id
        } msg]
        Log "Read status -- $::Settings($PlayerName,status) {$status} {$msg}"
        if {$::Settings($PlayerName,status) ne "Ready"} {
            Log "read failed -- $::Settings($PlayerName,status)"
            set index [lsearch -exact $::Settings(PlayerList) $PlayerName]
            if {$index != -1} {
                set ::Settings(PlayerList) [lreplace $::Settings(PlayerList) $index $index]
            }
            catch {chan close $::Settings($PlayerName,chan)}
            set index [lsearch -exact $::Settings(PlayerList) $PlayerName]
            if {$index != -1} {
                set ::Settings(PlayerList) [lreplace $::Settings(PlayerList) $index $index]
            }
        } else {
            Log "read done"
            return $::Settings($PlayerName,data)
        }
    } else {
        Log "no data expected "
        catch {chan puts $::Settings($PlayerName,channel) $data}
    }
    return
}

proc A {x}    { return [format %3d $x] }
proc P {name} { return [format %-15s \"$name\"] }
proc R {id}   { global Map ; return "[format %2d $id] [format %-25s \"$Map($id,name)\"]" }

proc Now {} {
    return [clock format [clock seconds] -gmt yes]
}

proc Log {text} {
    chan puts stdout "[Now] $text"
}

proc Display {text} {
    chan puts stderr $text
}

proc Comment {text} { Display "\# $text" }

##
## Send the bots what they need to know about maps
##
proc InitializeBotMaps {} {
    global Bonuses
    global Map

    # bonus-id (= super-region) --> #regions to pick.
    # standard picks.
    array set pickCount {
        1 1
        2 0
        3 1
        4 0
        5 2
        6 0
    }
    # random super-region picks
    foreach bonus [lrange [RandomizeList {1 3 4 5}] 0 1] {
        incr pickCount($bonus)
    }
    set pickList [list]
    foreach {bonus cnt} [array get pickCount] {
        lappend pickList {*}[lrange [RandomizeList $Bonuses($bonus,contains)] 0 $cnt]
    }
    set pickList [RandomizeList $pickList]
    set bonusInfo [list ]
    set regionBonusMap [list ]
    set regionInfo [list ]
    foreach bonus $::Settings(BonusList) {
        lappend bonusInfo $bonus $Bonuses($bonus,value)
    }
    foreach region $::Settings(RegionList) {
        lappend regionBonusMap $region $Map($region,bonus)
        if {[info exists Map($region,botNeighbors)] &&
	    [llength $Map($region,botNeighbors)]} {
            lappend regionInfo $region [join $Map($region,botNeighbors) {,}]
        }
    }
    foreach player $::Settings(PlayerList) {
        SendToPlayer $player "settings your_bot $player"
        foreach other $::Settings(PlayerList) {
            if {$other ne $player} {
                SendToPlayer $player "settings opponent_bot $other"
            }
        }
        catch {
            SendToPlayer $player "setup_map super_regions $bonusInfo"
            SendToPlayer $player "setup_map regions $regionBonusMap"
            SendToPlayer $player "setup_map neighbors [join $regionInfo { }]"
            SendToPlayer $player "pick_starting_regions $::Settings(Timeout) $pickList" yes 4
            set pickIndex($player) 0
        }
    }
    set playerList [RandomizeList $::Settings(PlayerList)]
    foreach i {1 2} {
        foreach player $playerList {
            while {$::Map([lindex $::Settings($player,data) $pickIndex($player)],ownedBy) ne "Neutral"} {
                incr pickIndex($player)
            }
            set region [lindex $::Settings($player,data) $pickIndex($player)]
            set ::Map($region,ownedBy) $player
            set ::Map($region,armyCount)  5
        }
    }

    return
}

##
## Normalize the bonus and map information
##
proc NormalizeMap {} {
    global Bonuses
    global Map

    foreach key [array names Bonuses *,contains] {
        set Bonuses($key) [lsort -integer $Bonuses($key)]
    }
    foreach region $::Settings(RegionList) {
        set Map($region,neighbors) [lsort -integer -unique $Map($region,neighbors)]
        foreach neighbor $Map($region,neighbors) {
            if {$region < $neighbor} {
                lappend Map($region,botNeighbors) $neighbor
            }
        }
        set Map($region,ownedBy) Neutral
        set Map($region,armyCount) 2
    }

    return
}

##
## Process a turn
##
proc TakeTurn {} {
    global Settings
    global Map

    # inGame: Array holding the set of players (still) holding
    # regions.

    Log     "Start of turn $Settings(Turn)."
    Display "turn $Settings(Turn)"
    Comment [string repeat . 60]

    ComputeBonusOwnership
    SendHistory
    SendMapViews
    foreach player $Settings(PlayerList) {
        GetAllocations $player
        GetMoves $player
    }
    set moveOrder [RandomizeList $Settings(PlayerList)]
    array unset Map *,preAttackCount
    foreach region $Settings(RegionList) {
        set Map($region,preAttackCount) $Map($region,armyCount)
    }
    ProcessMoves $moveOrder
    array set inGame {}
    foreach region $Settings(RegionList) {
        set inGame($Map($region,ownedBy)) 1
    }
    set playerCount 0
    foreach player $Settings(PlayerList) {
        if {![info exists inGame($player)]} {
            catch {chan close $Settings($player,chan)}
            Log     "Player $player has been eliminated."
            Display "[P $player] eliminated $::Settings(Turn)"
            set index [lsearch -exact $::Settings(PlayerList) $player]
            if {$index != -1} {
                set ::Settings(PlayerList) [lreplace $::Settings(PlayerList) $index $index]
            }
        }
    }
    Log "End of turn $Settings(Turn)."
    Comment [string repeat . 60]
    return
}

##
## See if any bonus are owned
##
proc ComputeBonusOwnership {} {
    global Bonuses
    global Map
    global Settings

    array unset Bonuses *,ownedBy
    foreach bonus $Settings(BonusList) {
        set owner $Map([lindex $Bonuses($bonus,contains) 0],ownedBy)
        if {$owner ne "Neutral"} {
            foreach region $Bonuses($bonus,contains) {
                if {$owner ne $Map($region,ownedBy)} {
                    set owner Neutral
                    break
                }
            }
        }
        set Bonuses($bonus,ownedBy) $owner
        if {$owner ne "Neutral"} {
	    Display "[P $owner] bonus $Bonuses($bonus,value) per $bonus <$Bonuses($bonus,name)>"
	}
    }

    Display {}
    return
}

##
##
##
proc GetAllocations {player} {
    global Settings
    global Map
    global Bonuses

    set armyAllocation 5
    foreach bonus $Settings(BonusList) {
        if {$player eq $Bonuses($bonus,ownedBy)} {
            incr armyAllocation $Bonuses($bonus,value)
        }
    }

    Display "[P $player] += $armyAllocation"

    set expectedCount $armyAllocation
    catch {SendToPlayer $player "settings starting_armies $armyAllocation" no}
    catch {SendToPlayer $player "go place_armies $Settings(Timeout)" yes}
    if {$Settings($player,status) eq "Ready"} {
        ##
        ## Verify that the response is of the correct one
        ##
        set errorDetected no
        foreach assignment [split $Settings($player,data) {,}] {
            set assignment [string trim $assignment]
            lassign $assignment name cmd region armies
            if {($name ne $player) || ($cmd ne "place_armies")} {
                Log "Invalid output received from $player, expected place_armies but received: {$Settings($player,data)}"
                catch {chan close $Settings($player,chan)}
                set index [lsearch -exact $::Settings(PlayerList) $player]
                if {$index != -1} {
                    set ::Settings(PlayerList) [lreplace $::Settings(PlayerList) $index $index]
                }
                set errorDetected yes
                break
            }
            if {$Map($region,ownedBy) ne $player} {
                Display "<$player> invalid !owned [R $region] held $Map($region,ownedBy)"
                Log "Invalid assigment received from $player, attempt to assign armies to $region ($Map($region,name)) owned by '$Map($region,ownedBy)'"
                #catch {chan close $Settings($player,chan)}
                #set index [lsearch -exact $::Settings(PlayerList) $player]
                #if {$index != -1} {
                #    set ::Settings(PlayerList) [lreplace $::Settings(PlayerList) $index $index]
                #}
                #set errorDetected yes
                #break
            }
            incr armyAllocation -$armies
            incr Map($region,armyCount) $armies
            AddPlacementToHistory $player $region $armies
        }
        if {!$errorDetected && $armyAllocation} {
            set armyAllocation [expr {$expectedCount - $armyAllocation}]
            Display "<$player> invalid bad-number expected $expectedCount received $armyAllocation"
            Log "Invalid assigment received from $player, should have assigned $expectedCount, but received $armyAllocation"
            catch {chan close $Settings($player,chan)}
            set index [lsearch -exact $::Settings(PlayerList) $player]
            if {$index != -1} {
                set ::Settings(PlayerList) [lreplace $::Settings(PlayerList) $index $index]
            }
        }
    }

    Display {}
    return
}

##
##
##
proc GetMoves {player} {
    global Settings
    global Map

    set Settings($player,moves) [list]
    catch {SendToPlayer $player "go attack/transfer $Settings(Timeout)" yes}
    if {($Settings($player,status) eq "Ready") && ($Settings($player,data) ne "No moves")} {
        ##
        ## Verify that the response is of the correct one
        ##
        set errorDetected no
        foreach movement [split $Settings($player,data) {,}] {
            set movement [string trim $movement]
            lassign $movement name cmd fromRegion toRegion armies
            if {($name ne $player) || ($cmd ne "attack/transfer")} {
                Log "Invalid output received from $player, expected attack/transfer but received: {$Settings($player,data)}"
                catch {chan close $Settings($player,chan)}
                set index [lsearch -exact $::Settings(PlayerList) $player]
                if {$index != -1} {
                    set ::Settings(PlayerList) [lreplace $::Settings(PlayerList) $index $index]
                }
                set errorDetected yes
                break
            }
            if {$Map($fromRegion,ownedBy) ne $player} {
                #chan puts stderr "Invalid attack/transfer received from $player, attempt to attack from $fromRegion ($Map($fromRegion,name)) owned by '$Map($fromRegion,ownedBy)'"
                #chan puts stdout "Invalid attack/transfer received from $player, attempt to attack from $fromRegion ($Map($fromRegion,name)) owned by '$Map($fromRegion,ownedBy)'"
                #catch {chan close $Settings($player,chan)}
                #set index [lsearch -exact $::Settings(PlayerList) $player]
                #if {$index != -1} {
                #    set ::Settings(PlayerList) [lreplace $::Settings(PlayerList) $index $index]
                #}
                #set errorDetected yes
                #break
            }
            lappend Settings($player,moves) [list $fromRegion $toRegion $armies]
        }
    }
    return;
}

##
## Send the players what they can see of the maps
##
proc SendMapViews {} {
    global Settings
    global Map
    global Bonuses

    foreach player $Settings(PlayerList) {
        array unset viewable
        set armyAllocation 5
        foreach region $Settings(RegionList) {
            if {$player eq $Map($region,ownedBy)} {
                set viewable($region) 1
                foreach neighbor $Map($region,neighbors) {
                    set viewable($neighbor) 1
                }
            }
        }
        if {![array size viewable]} {
            continue
        }
        set cmd update_map
	set table {}
        foreach region [array names viewable] {
            append cmd " " $region " " $Map($region,ownedBy) " " $Map($region,armyCount)

	    lappend table [list $region $Map($region,ownedBy) $Map($region,armyCount)]
        }
        catch {SendToPlayer $player $cmd}

	# Write what a player knows of the map to stdout, in a readable format.
	foreach item [lsort -dict -index 1 [lsort -integer -index 2 $table]] {
	    lassign $item rid owner armies
	    Display "[P $player] view [R $rid] [P $owner] [A $armies]"
	}
	Display {}
    }

    return;
}

##
## Process a move set
##
proc ProcessMoves {playerList} {
    global Map
    global Settings

    set moveMax 0
    foreach player $playerList {
        set moveMax [expr {max($moveMax,[llength $Settings($player,moves)])}]
    }
    for {set moveNumber 0} {$moveNumber <= $moveMax} {incr moveNumber} {
        foreach player $playerList {
            set move [lindex $Settings($player,moves) $moveNumber]
            if {[llength $move]} {
                lassign $move fromRegion toRegion armyCount
                if {($Map($fromRegion,ownedBy) eq $player) &&
                    $Map($fromRegion,preAttackCount) &&
                    ($Map($fromRegion,armyCount) > 1)} {

                    set armiesToMove [expr {min($armyCount,$Map($fromRegion,preAttackCount)-1,$Map($fromRegion,armyCount)-1)}]

                    AddAttackToHistory $player $fromRegion $toRegion $armiesToMove

                    if {$Map($toRegion,ownedBy) eq $player} {
			# Transfer. Simply record.
                        incr Map($toRegion,armyCount) $armiesToMove
                        set Map($fromRegion,preAttackCount) \
                            [expr {max(0,$Map($fromRegion,preAttackCount) - $armiesToMove)}]
                        set Map($fromRegion,armyCount) \
                            [expr {max(1,$Map($fromRegion,armyCount) - $armiesToMove)}]
                    } else {
			# Attack.
                        set defendingArmies $Map($toRegion,armyCount)

			# compute kill ratios.
                        set offensiveKills [expr {int(0.60 * $armiesToMove)}]
                        set defensiveKills [expr {int(0.70 * $defendingArmies)}]

			Display "[P $player] versus [A $Map($fromRegion,preAttackCount)] - [A $defendingArmies]"
			Display "[P $player] fight  [A $armiesToMove] - [A $defendingArmies]"
			Display "[P $player] kills  [A $offensiveKills] - [A $defensiveKills]"

			# Left in attack source.
                        set Map($fromRegion,preAttackCount) \
                            [expr {max(0,$Map($fromRegion,preAttackCount) - $defensiveKills)}]
                        set Map($fromRegion,armyCount) \
                            [expr {max(1,$Map($fromRegion,armyCount) - $defensiveKills)}]

			# Left to move/defend
                        set armiesToMove \
                            [expr {max(0,$armiesToMove - $defensiveKills)}]
                        set defenders \
                            [expr {$Map($toRegion,armyCount) - $offensiveKills}]

			Display "[P $player] left   [A $armiesToMove] - [A $defenders]"

                        if {!$armiesToMove} {
			    # Attack failed, no armies left to move in.
                            set Map($toRegion,armyCount) [expr {max(1,$defenders)}]
			    Display "[P $player] failed exhausted"
                        } elseif {$defenders > 0} {
			    # Attack failed, defenders prevailed.
                            set Map($toRegion,armyCount) [expr {max(1,$defenders)}]
			    Display "[P $player] failed"
                        } else {
			    Display "[P $player] success"
			    Display "[P $player] move   [A $armiesToMove]"

                            set Map($toRegion,armyCount) $armiesToMove
                            set Map($toRegion,preAttackCount) 0
                            set Map($toRegion,ownedBy) $player

                            set Map($fromRegion,preAttackCount) \
                                [expr {max(0,$Map($fromRegion,preAttackCount) - $armiesToMove)}]
                            set Map($fromRegion,armyCount) \
                                [expr {max(1,$Map($fromRegion,armyCount) - $armiesToMove)}]
                        }

			Display "[P $player] final  [A $Map($fromRegion,armyCount)] - [A $Map($toRegion,armyCount)]"
			Display {}
                    }
                }
            }
        }
    }
    return
}

##
## Add an attack to what the neightbors can see
##
proc AddAttackToHistory {PlayerName FromRegion ToRegion ArmyCount} {
    global History
    global Map

    set turn $::Settings(Turn)
    set ownedBy $Map($ToRegion,ownedBy)
    set move [list $PlayerName attack/transfer $FromRegion $ToRegion $ArmyCount]

    Display "[P $PlayerName] [expr {($PlayerName eq $ownedBy) ? "transfer":"attack  "}] [R $FromRegion] [R $ToRegion] [A $ArmyCount]"

    # Tell attacked player (destination) about the move.
    if {($ownedBy ne "Neutral") && ($ownedBy ne $PlayerName)} {
        set History($PlayerName,$turn,$move) $move
    }
    # Tell other players, neighbours of the source region.
    foreach neighbor $Map($FromRegion,neighbors) {
        set ownedBy $Map($neighbor,ownedBy)
        if {($ownedBy ne "Neutral") && ($ownedBy ne $PlayerName)} {
            set History($ownedBy,$turn,$move) $move
        }
    }
    # Tell other players, neighbours of the destination region.
    foreach neighbor $Map($ToRegion,neighbors) {
        set ownedBy $Map($neighbor,ownedBy)
        if {($ownedBy ne "Neutral") && ($ownedBy ne $PlayerName)} {
            set History($ownedBy,$turn,$move) $move
        }
    }

    return
}

##
## Add an placement to what the neightbors can see
##
proc AddPlacementToHistory {PlayerName ToRegion ArmyCount} {
    global History
    global Map

    set turn $::Settings(Turn)
    set move [list $PlayerName place_armies $ToRegion $ArmyCount]

    Display "[P $PlayerName] place [R $ToRegion] [A $ArmyCount]"

    foreach neighbor $Map($ToRegion,neighbors) {
        set ownedBy $Map($neighbor,ownedBy)
        if {($ownedBy ne "Neutral") && ($ownedBy ne $PlayerName)} {
            set History($ownedBy,$turn,$move) $move
        }
    }

    return
}

##
## Send the history of what the players saw last turn
##
proc SendHistory {} {
    global History

    Comment "TODO: show history"

    set turn [expr {$::Settings(Turn) - 1}]
    foreach player $::Settings(PlayerList) {
        set moveList [list]
        set moveKeyList [array names History $player,$turn,*]
        if {[llength $moveKeyList]} {
            foreach key $moveKeyList {
                lappend moveList $History($key)
            }
            SendToPlayer $player "opponent_moves [join $moveList {, }]"
        } else {
            Log "No moves for $player on turn $turn"
        }
    }

    return
}


##      GameName                - name of game
##      GameBaseDirectory       - path (absolute or relative to current directory)
##      MapSpecificationFile    - Specification file (absolute or relative to
##                                GameBaseDirectory) for map
##      <playerInfo> -- conists of pairs of argument
##          PlayerName      - name of player
##          PlayerBotFile   - path (absolute or relative to GameBaseDirectory)
##                            for bot code
lassign $argv GameName GameBaseDirectory MapSpecificationFile
cd $GameBaseDirectory
Log "Game: $GameName"
if {[catch {source $MapSpecificationFile} msg]} {
    chan puts stderr "Map definition failed due to: $msg"
    chan puts stderr $errorInfo
    exit 1
}
foreach {PlayerName PlayerBotFile} [lrange $argv 3 end] {
    if {[catch {LaunchBot $PlayerName $PlayerBotFile} msg]} {
        chan puts stderr [format {Player '%1$s' failed to launch '%2$s' due to: %3$s} $PlayerName $PlayerBotFile $msg]
    } else {
        Log [format {%1$s: "%2$s"} PlayerName $PlayerBotFile]
    }
}
if {![llength $Settings(PlayerList)]} {
    chan puts stderr "All bots failed to initialize"
    exit 1
}
NormalizeMap
InitializeBotMaps

Log "Start of game."
while {([llength $Settings(PlayerList)] > 1) && ($::Settings(Turn) < $::Settings(MaxTurn) )} {
    incr ::Settings(Turn)
    TakeTurn
    Log [string repeat {*} 80]
}
if {[llength $Settings(PlayerList)] > 1} {
    Log    "Tie."
    Display tie
} else {
    Log     "Winner is $Settings(PlayerList) in $::Settings(Turn) turns."
    Display "winner $::Settings(Turn) [P [lindex $Settings(PlayerList) 0]]"
}
Log "End of game."
Display end

exit 0
